#include "platform.h"
#include "memmap.h"
#include "crc.h"
#include "bootloader.h"
#include "slip.h"
#include "serial_bridge.h"
#include "cm3_scb_registers.h"
void bootloader_task(void *arg);

// Because changing VTOR in ISR causes a Hardfault
volatile uint16_t b_jump_to_user_code = 0;

extern uint32_t __fastcode_load;
extern uint32_t __fastcode_start;
extern uint32_t __fastcode_end;

xSemaphoreHandle xWaitForBootloaderTriggerSemaphore = NULL;
xSemaphoreHandle xWaitForBootloaderTransferSemaphore = NULL;

// Time to wait for a bootloader trigger. Set to portMAX_DELAY to make it infinite.
const portTickType xBootloaderTriggerWaitDelay = 5000 / portTICK_RATE_MS;

// core reset
void system_reset();

int main(void)
{
  // Copy (.fastcode) functions to RAM (.fastcode)
  uint32_t* src = &__fastcode_load;
  uint32_t* dest = &__fastcode_start;
  while (dest < &__fastcode_end)
  {
    *(dest++) = *(src++);
  }

  platform_init();

  xTaskCreate(bootloader_task, (const signed char * const)"bootloader",  configMINIMAL_STACK_SIZE*4, NULL, 1, NULL);

  platform_run();
  return 0;
}

void init_bootloader()
{
  crc_enable();

  slip_init(serial_bridge_process_packet);
  uart_set_rx_handler(uart_print, slip_rx_char, NULL);
}
void bootloader_task(void *arg)
{
  vSemaphoreCreateBinary(xWaitForBootloaderTriggerSemaphore);
  vSemaphoreCreateBinary(xWaitForBootloaderTransferSemaphore);

  // Take semaphores because they start 'Given'
  xSemaphoreTake(xWaitForBootloaderTriggerSemaphore, 0);
  xSemaphoreTake(xWaitForBootloaderTransferSemaphore, 0);

  init_bootloader();

  if(xWaitForBootloaderTriggerSemaphore != NULL)
  {
    leds_on(1);   // Green
  
    // Wait for bootloader trigger
    if (xSemaphoreTake(xWaitForBootloaderTriggerSemaphore, xBootloaderTriggerWaitDelay) == pdTRUE)
    {
      leds_off(0xff);
      leds_on(2);   // Red
      // Block until image transfer is done.
      xSemaphoreTake(xWaitForBootloaderTransferSemaphore, portMAX_DELAY);
    }
system_reset();
    { 
      volatile uint32_t usr_boot_addr;
      void (*user_app)(void) = NULL;

      __asm__("cpsid i");
      *(mem_get_reg32(CM3_SCB_BASE_ADDRESS + CM3_SCB_VTOR_OFFSET)) = (uint32_t)(user_code_base_address);

      // Jump to user application
      usr_boot_addr = *(volatile uint32_t*) (user_code_base_address + 4); 
      user_app = (void (*)(void)) usr_boot_addr;
      user_app();
    }
  }
  while(1);
}

void system_reset()
{
  *cm3_scb_get_AIRCR() = 0x05FA0001; // Request a VECTRESET : Only core is reset !
  __asm("DSB \n");  // wait for flash memory op to complete
  while (1);
}

